#!/usr/bin/env python
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
 
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.1"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"
 
###############################################################################

# System imports
import logging
import subprocess
import os 
import pickle

from databases import Databases
from matrix_generator import MatrixGenerator
from gff_generator import GffGenerator
from genome import Genome, AnnotationParser

###############################################################################
################################ - Classes - ##################################

class Annotate:

    GENOME_BIN          = 'genome_bin'
    GENOME_PROTEINS     = 'genome_proteins'    
    GENOME_KO           = 'annotations_ko'
    GENOME_COG          = 'annotations_cog'
    GENOME_PFAM         = 'annotations_pfam'
    GENOME_TIGRFAM      = 'annotations_tigrfam'
    GENOME_GFF          = 'annotations_gff'
    GENOME_OBJ          = 'annotations_genomes'
    OUTPUT_KO           = 'ko_frequency_table.tsv'
    OUTPUT_PFAM         = 'pfam_frequency_table.tsv'
    OUTPUT_TIGRFAM      = 'tigrfam_frequency_table.tsv'

    GFF_SUFFIX          = '.gff'
    PROTEINS_SUFFIX     = '.faa'
    ANNOTATION_SUFFIX   = '.tsv'
    PICKLE_SUFFIX       = '.pickle'
    
    def __init__(self,
                 output_directory,
                 ko,
                 pfam,
                 tigrfam,
                 cog,
                 evalue,
                 bit,
                 id,
                 aln_query,
                 aln_reference,
                 threads,
                 suffix):

        # Define inputs and outputs
        self.output_directory = output_directory
        self.gff_directory    = os.path.join(output_directory, )

        # Define type of annotation to be carried out
        self.ko               = ko 
        self.pfam             = pfam 
        self.tigrfam          = tigrfam 
        self.cog              = cog 
        # Cutoffs
        self.evalue           = evalue 
        self.bit              = bit 
        self.id               = id 
        self.aln_query        = aln_query
        self.aln_reference    = aln_reference
        # Parameters
        self.threads          = threads
        self.suffix           = suffix

        # Load databases
        self.databases        = Databases()

    def prep_genome(self, genome_file_list):
        '''
        Do any preparation specific to the genome annotation pipeline. 

        Inputs
        ------
        genome_file_list - List. list of strings, each a path to a file
        containing a genome 

        Outputs
        -------
        returns the directory with all genome ids sym-linked into it.
        '''
        # link all the genomes into one file    
        genome_directory=None
        if genome_file_list:
            genome_directory = os.path.join(self.output_directory, 
                                           self.GENOME_BIN)
            os.mkdir(os.path.join(self.output_directory, self.GENOME_BIN))
            for genome_path in genome_file_list:
                os.symlink(os.path.join(os.getcwd(),
                                        genome_path), 
                           os.path.join(genome_directory, 
                                        os.path.basename(genome_path)
                                        )
                           )    
        return genome_directory


    def call_proteins(self, genome_directory):
        '''
        Use prodigal (https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-11-119)
        to call proteins within the genomes 
        
        Parameters
        ----------
        genome_directory  - string. Directory containing .fna files for each  
                            input genome

        Outputs
        -------
        returns the directory containing an .faa file for each input genomes
        '''   
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_PROTEINS) 
        os.mkdir(output_directory_path)
        genomes_list = []

        for genome in os.listdir(genome_directory):
            if genome.endswith(self.suffix):
                genome_path = os.path.join(genome_directory,
                                           genome)
                output_genome = os.path.splitext(genome)[0] + self.PROTEINS_SUFFIX 
                output_genome_path = os.path.join(output_directory_path,
                                                  output_genome)
                logging.info("    - Calling proteins for genome: %s" % (genome))
                cmd = 'prodigal -q -p meta -o /dev/null -a %s -i %s' % (output_genome_path,
                                                        genome_path)
                
                logging.debug(cmd)
                subprocess.call(cmd, shell = True)
                genome = Genome(output_genome_path)
                genomes_list.append(genome)
        return genomes_list
    
    def annotate_ko(self, genomes_list):
        '''
        Annotate the proteins encoded by each genome with KO ids using either BLAST or using HMM
        searches (no implemented yet).

        Parameters
        ----------        
        genome_faa_directory  - string. Directory containing .faa files for 
                                each input genome
        
        Outputs
        -------
        returns a directory containing the search results for each of the input population genomes, 
        and a frequency matrix contining with the KOs as rows, and the genomes as columns.
        '''        

        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_KO)
        os.mkdir(output_directory_path)
        for genome in genomes_list:
            output_annotation_path = os.path.join(output_directory_path, genome.name) + self.ANNOTATION_SUFFIX
            self._diamond_search(genome.path, output_annotation_path, self.databases.KO_DB)
            genome.add(output_annotation_path, 
                         self.evalue, 
                         self.bit, 
                         self.aln_query, 
                         self.aln_reference,
                         AnnotationParser.KO)

    def annotate_cog(self): 
        pass
        # TODO haven't prepared the reference database yet

    def _diamond_search(self, input_genome_path, output_path, database):
        '''
        Carry out a diamond blastp search. 

        Parameters
        ----------   
        input_genome_path     - string. Path to file containing .faa file for 
                                an input genome
        output_path           - string. Path to file to output results into    
        databases             - string. Path to HMM to use for searching         
        '''  
         
        cmd = 'diamond blastp --quiet --outfmt 6 --max-target-seqs 1 --query %s --out %s --db %s --threads %s ' \
                            % (input_genome_path,  output_path, database, self.threads)
        if self.evalue:
            cmd += '--evalue %f ' % (self.evalue) 
        if self.bit:
            cmd += '--min-score %f ' % (self.bit)
        if self.id:
            cmd += '--id %f ' % (self.id)
        logging.debug(cmd)
        subprocess.call(cmd, shell = True)

    def annotate_pfam(self, genomes_list):
        '''
        Annotate the proteins encoded by each genome with pfam ids using HMM searches.

        Parameters
        ----------        
        genomes_list  - list. list of Genome objects        

        '''    
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_PFAM)
        os.mkdir(output_directory_path)
        for genome in genomes_list:
            output_annotation_path = os.path.join(output_directory_path, genome.name) + self.ANNOTATION_SUFFIX
            self._hmm_search(genome.path, output_annotation_path, self.databases.PFAM_DB)
            genome.add(output_annotation_path, 
                         self.evalue, 
                         self.bit, 
                         self.aln_query, 
                         self.aln_reference,
                         AnnotationParser.PFAM)
    ### ~ TODO: arCOG
    
    def annotate_tigrfam(self, genomes_list):
        '''
        Annotate the proteins encoded by each genome with tigrfam ids using HMM searches.

        Parameters
        ----------        
        genomes_list  - list. list of Genome objects        

        '''    
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_TIGRFAM)
        os.mkdir(output_directory_path)      
        for genome in genomes_list:
            output_annotation_path = os.path.join(output_directory_path, genome.name) + self.ANNOTATION_SUFFIX
            self._hmm_search(genome.path, output_annotation_path, self.databases.TIGRFAM_DB)
            genome.add(output_annotation_path, 
                         self.evalue, 
                         self.bit, 
                         self.aln_query, 
                         self.aln_reference,
                         AnnotationParser.TIGRFAM)

    def _hmm_search(self, input_genome_path, output_path, database):
        '''
        Carry out a hmmsearch. 

        Parameters
        ----------   
        input_genome_path     - string. Path to file containing .faa file for 
                                an input genome
        output_path           - string. Path to file to output results into   
        databases             - string. Path to HMM to use for searching          
        '''  
        cmd = "hmmsearch --cpu %s -o /dev/null --noali --domtblout %s " \
                          % (self.threads, output_path)
        if self.evalue:
            cmd += '-E %f ' % (self.evalue) 
        if self.bit:
            cmd += '-T %f ' % (self.bit)
        if self.id:
            logging.warning("--id flag not used for hmmsearch")
        cmd += "%s %s " % (database, input_genome_path)
        logging.debug(cmd)
        subprocess.call(cmd, shell = True)        

    def _parse_genome_proteins_directory(self, directory):
        '''
        Iterate through a directory and parse all .faa files it contains (assumed to be separate)
        genome bins

        Parameters 
        ----------
        directory   -   string. path to directory containing genome proteins.

        Outputs
        -------
        A list of Genome objects.
        '''
        genomes_list = []
        for genome_proteins_file in os.listdir(directory):
            if genome_proteins_file.endswith(self.suffix):
                genome = Genome(os.path.join(directory, genome_proteins_file))
                genomes_list.append(genome) 
        return genomes_list

    def _generate_gff_files(self, genomes_list):
        '''
        Write GFF files for each of the genome objects in genomes_list

        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = os.path.join(self.output_directory, 
                                             self.GENOME_GFF)
        os.mkdir(output_directory_path)      
        for genome in genomes_list:
            logging.info('    - Generating .gff file for %s' % genome.name)
            gff_output = os.path.join(output_directory_path, genome.name + self.GFF_SUFFIX)
            gg = GffGenerator()
            gg.write(genome, gff_output)

    def _rename_fasta(self, genomes_list):
        '''
        Rename the called proteins with annotation ids.
        
        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        
        for genome in genomes_list:
            with open(genome.path, 'w') as genome_fasta_io:
                for sequence_name in genome.protein_ordered_dict.values():
                    annotations = ' '.join(genome.sequences[sequence_name].all_annotations())
                    genome_fasta_io.write( ">%s %s\n" % (sequence_name, annotations) )
                    genome_fasta_io.write( genome.sequences[sequence_name].seq + '\n' )

    def _pickle_objects(self, genomes_list):
        '''
        Store annotated genome objects as pickles.
        
        Parameters
        ----------
        genomes_list - List. List of Genome objects
        '''
        output_directory_path = os.path.join(self.output_directory,
                                             self.GENOME_OBJ)
        os.mkdir(output_directory_path)
        for genome in genomes_list:
            with open(os.path.join(output_directory_path, genome.name + self.PICKLE_SUFFIX), 'w') as output:
                pickle.dump(genome, output)

    def do(self, genome_directory, protein_directory, genome_files, protein_files):
        '''
        Run Annotate pipeline for enrichM

        Parameters
        ----------
        genome_directory    - String. Path to directory containing genomes
        protein_directory   - String. Path to directory containing proteins (.faa files) for genomes
        genome_files        - List. List of strings, each to a .fna genome file.
        protein_files       - List. List of strings, each to a .faa proteins file.
        '''

        logging.info("Running pipeline: annotate")
        logging.info("Setting up for genome annotation")

        if protein_directory:
            logging.info("Using provided proteins.")
            genomes_list = self._parse_genome_proteins_directory(protein_directory)
        elif protein_files:
            logging.info("Using provided proteins.")
            genomes_list = [Genome(protein_file) for protein_file in protein_files]
        elif genome_directory:
            logging.info("Calling proteins for annotation")
            genomes_list = self.call_proteins(genome_directory)
        elif genome_files:
            logging.info("Calling proteins for annotation")            
            genomes_list = self.call_proteins(self.prep_genome(genome_files))

        if len(genomes_list)==0:
            logging.error('There were no genomes found with the suffix %s within the provided directory' \
                                        %  (self.suffix))
        else:
            logging.info("Starting annotation:")
            if self.ko:
                logging.info('    - Annotating genomes with ko ids')
                self.annotate_ko(genomes_list)

                logging.info('    - Generating ko frequency table')
                mg = MatrixGenerator(MatrixGenerator.KO)
                
                freq_table = os.path.join(self.output_directory, self.OUTPUT_KO)
                mg.write_matrix(genomes_list, freq_table)

            if self.cog:
                logging.info('    - Annotating genomes with COG ids')
                logging.info('    - COG annotation currently not implemented')

            if self.pfam:
                logging.info('    - Annotating genomes with pfam ids')
                self.annotate_pfam(genomes_list)

                logging.info('    - Generating pfam frequency table')
                mg = MatrixGenerator(MatrixGenerator.PFAM)
                
                freq_table = os.path.join(self.output_directory, self.OUTPUT_PFAM)
                mg.write_matrix(genomes_list, freq_table)

            if self.tigrfam:
                logging.info('    - Annotating genomes with tigrfam ids')
                self.annotate_tigrfam(genomes_list)
                
                logging.info('    - Generating tigrfam frequency table')
                mg = MatrixGenerator(MatrixGenerator.TIGRFAM)
                
                freq_table = os.path.join(self.output_directory, self.OUTPUT_TIGRFAM)
                mg.write_matrix(genomes_list, freq_table)

            logging.info('Generating .gff files:')
            self._generate_gff_files(genomes_list)

            logging.info('Renaming protein headers')
            self._rename_fasta(genomes_list)

            logging.info('Storing genome objects')
            self._pickle_objects(genomes_list)

            logging.info('Finished annotation')




#!/usr/bin/env python
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
 
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.1"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"
 
###############################################################################

import logging
import pickle
import os
import random
import itertools

from traverse import NetworkTraverser
from databases import Databases


###############################################################################

class NetworkBuilder:
       
    MODULE_PREFIX   = 'M'
    COMPOUND_PREFIX = 'C'    
    MAP_PREFIX      = 'map'
    PATHWAY_PREFIX  = 'rn'
    REACTION_PREFIX = 'R'
    ZERO            = '0.0'

    def __init__(self, metadata_keys):
        
        self.d=Databases()

        self.metadata_keys \
                        = metadata_keys
        self.matrix_header \
                        = ["compound", "reaction"] + \
                          [key + '_reaction_metagenome' 
                           for key in self.metadata_keys]
        self.transcriptome_header \
                        = [key + '_reaction_transcriptome' 
                           for key in self.metadata_keys] + \
                          [key + '_reaction_expression' 
                           for key in self.metadata_keys]
        self.compound_header \
                        = [key + '_compound' 
                           for key in self.metadata_keys]
        self.metadata_header \
                        = ['node', 
                           'description',
                           'type',
                           'module', 
                           'module_descr',
                           'pathway', 
                           'pathway_descr', 
                           'node_type']
        self.query_header \
                        =  ['query',
                            'step']
        self.compound_reaction_index_header \
                        = []
        self.step_header = ['step']
    def _gather_module(self, key):
        if key in self.d.r2m:
            len_list =  [len(self.d.m2r[x]) for x in self.d.r2m[key]]
            module = self.d.r2m[key][len_list.index(max(len_list))]
            module_description = self.d.m[module]
        else:
            module = 'NA'
            module_description='NA'
        return module, module_description
    
    def _gather_pathway(self, key):
        if key in self.d.r2p:
            len_list =  [len(self.d.p2r[x]) for x in self.d.r2p[key]]
            pathway = self.d.r2p[key][len_list.index(max(len_list))]
            pathway_description = self.d.p[pathway]
        else:
            pathway = 'NA'
            pathway_description='NA'
        return pathway, pathway_description
    
    def _parse_queries(self, queries_file):
        '''
        Parse file with one column, each entry a compound ID from kegg
        Parameters
        ----------
        queries_file    - String. Path to file containing Compound ids 
        
        Output
        ------
        Set. Compound Ids to start from in the explore network
        '''
        return set([x.strip() for x in open(queries_file)])


    def normalise(self, reaction_abundances):
        probability_list = []
        for reaction_abundance in reaction_abundances:
            probability = float(reaction_abundance)/sum(reaction_abundances)
            probability_list.append(probability)
        return probability_list

    def get_transition(self, probabilities):
        n=random.uniform(0,1)
        x = []
        for probability in probabilities:
            x.append(probability-n)
        return x.index(min(x, key=abs))

    def traverse_matrix(self,
                        abundances_metagenome,
                        abundances_transcriptome,
                        possible_reactions,
                        query_list,
                        number_of_queries,
                        steps): 

        network_visitations={key:{} for key in self.metadata_keys}
        for _ in range(number_of_queries):
            starting_compound = random.choice(query_list)
            previous_reaction=''
            next_compound=''
            for i in range(steps):
                if i==0:
                    iter_batch={key:starting_compound for key in self.metadata_keys}

                for key in self.metadata_keys:
                    compound = iter_batch[key]
                    
                    try:
                        reactions = [x for x in self.d.c2r[compound]
                                     if x in self.d.r2rpair]
                    except:
                        reactions=[]
                    possible_reaction_abundance = []
                    possible_reaction_name = []
                    for reaction in reactions:  
                        if(reaction in abundances_metagenome[key] and
                           reaction in abundances_transcriptome[key]):
                            if reaction!=previous_reaction:
                                if abundances_metagenome[key][reaction]>0:
                                    possible_reaction_name.append(reaction)
                                    possible_reaction_abundance.append(abundances_transcriptome[key][reaction])

                    if any(possible_reaction_abundance):
                        probabilities = self.normalise(possible_reaction_abundance)
                        transition_reaction = possible_reaction_name[self.get_transition(probabilities)]
                        if any(self.d.r2rpair[transition_reaction]):
                            for pair in self.d.r2rpair[transition_reaction]:
                                if compound in pair.split('_'):
                                    transition_compound = pair.replace(compound, '')\
                                                              .replace('_', '')   
                            if transition_compound not in network_visitations[key]:
                                network_visitations[key][transition_compound]=1
                            else:
                                network_visitations[key][transition_compound]+=1    
                            iter_batch[key] = transition_compound
                            previous_reaction=transition_reaction
                            
        output_lines = ['\t'.join(['C'] + self.metadata_keys)]
        for c in set(itertools.chain(*possible_reactions.values())):
            output_line = [c]
            for key in self.metadata_keys:
                if c in network_visitations[key]:
                    output_line.append(str(network_visitations[key][c]))
                else:
                    output_line.append('0')    
            output_lines.append('\t'.join(output_line))
        return output_lines

    def all_matrix(self, 
                   abundances_metagenome, 
                   abundances_transcriptome,
                   abundances_expression,
                   abundances_metabolome,
                   reference_dict):
        '''
        Parameters
        ----------
        '''
        seen_nodes = set()
        # Construct headers to network matrices
        if(abundances_transcriptome and abundances_expression):
            network_lines  = ['\t'.join(self.matrix_header + 
                                        self.transcriptome_header)]
        else:
            network_lines  = ['\t'.join(self.matrix_header)]  
        if abundances_metabolome:
            node_metadata_lines = ['\t'.join(self.metadata_header + 
                                             self.compound_header)]
        else:     
            node_metadata_lines = ['\t'.join(self.metadata_header)]
        
        for reaction, entry in reference_dict.items():
            for compound in entry:
                if any([reaction in x.keys() 
                        for x in abundances_metagenome.values()]):
                    #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                    #~#~#~#~#~#~#~#~#~#~ Fill in abundances ~#~#~#~#~#~#~#~#~#~
                    reaction_line = [compound, reaction]

                    for key in self.metadata_keys:
                        if reaction in abundances_metagenome[key]:
                            reaction_line.append(str(abundances_metagenome[key][reaction]))
                        else:
                            reaction_line.append(self.ZERO)

                    if(abundances_transcriptome and abundances_expression):
                        for key in self.metadata_keys:
                            if reaction in abundances_transcriptome[key]:
                                reaction_line.append(str(abundances_transcriptome[key][reaction]))
                            else:
                                reaction_line.append(self.ZERO)

                        for key in self.metadata_keys:
                            if reaction in abundances_expression[key]:
                                reaction_line.append(str(abundances_expression[key][reaction]))
                            else:
                                reaction_line.append(self.ZERO)

                    output_line = '\t'.join(reaction_line)
                    if output_line not in network_lines:
                        network_lines.append(output_line)
                        
                    #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                    #~#~#~#~#~#~#~#~#~ Fill in node metadata ~#~#~#~#~#~#~#~#~#                    
                    if compound not in seen_nodes:
                        if compound in self.d.compound_desc_dict:
                            compound_type = \
                                ','.join(self.d.compound_desc_dict[compound]['A'])
                        else:
                            compound_type = 'NA'
                        
                        metadata_list = [compound, self.d.c[compound], compound_type, 
                                         'NA', 'NA', 'NA', 'NA', 'compound']
                        if abundances_metabolome:   
                            # This is fragile - compound abundances need to be aggregated
                            # by the user. TODO: Integrate compounds into kegg_matrix
                            # module.
                            for sample in self.metadata_keys:
                                c_ab = abundances_metabolome.get_entry(sample, compound)
                                if c_ab == 0:
                                    c_ab = '-5'
                                metadata_list.append(str(c_ab))
                        node_metadata_lines.append('\t'.join(metadata_list))
                        seen_nodes.add(compound)

                    if reaction not in seen_nodes:
                        module, module_description \
                                        = self._gather_module(reaction)
                        pathway, pathway_description \
                                        = self._gather_pathway(reaction)
                        metadata_list = [reaction, 
                                         self.d.r[reaction],
                                         'NA', 
                                         module, 
                                         module_description, 
                                         pathway,
                                         pathway_description,
                                         'reaction']
                        if abundances_metabolome:   
                            metadata_list += ['-5' for sample in self.metadata_keys]
                        node_metadata_lines.append('\t'.join(metadata_list))
                        seen_nodes.add(reaction)

        return network_lines, node_metadata_lines

    def query_matrix(self, 
                     abundances_metagenome, 
                     abundances_transcriptome,
                     abundances_expression,
                     queries, 
                     depth):
        '''
        Parameters
        ----------
        '''
        steps         = 0
        queries_list  = self._parse_queries(queries)
        seen_steps    = set()
        seen_nodes    = set()
        level_queries = set()

        if(abundances_transcriptome and abundances_expression):
            network_lines  \
                    = ['\t'.join(self.matrix_header + 
                                 self.transcriptome_header +
                                 self.step_header)]
        else:
            network_lines  \
                    = ['\t'.join(self.matrix_header+
                                 self.step_header)] 
        node_metadata_lines \
                    = ['\t'.join(self.metadata_header + 
                                 self.query_header )]
        compound_reaction_index_lines \
                    = ['\t'.join(self.compound_reaction_index_header)]
        while depth>0:
            if any(level_queries):
                queries_list = set(level_queries)
                level_queries = set()
            for reaction, entry in self.d.r2c.items():
                if any([reaction in x.keys() 
                        for x in abundances_metagenome.values()]):

                    if any(queries_list.intersection(entry)):
                        reaction_compounds = [x for x in entry 
                                              if x not in NetworkTraverser.to_omit]
                        for compound in reaction_compounds:
                            
                            #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                            #~#~#~#~#~#~#~#~ Fill in abundances ~#~#~#~#~#~#~#~
                            reaction_line = [compound, reaction]
                        
                            for key in self.metadata_keys:
                                if reaction in abundances_metagenome[key]:
                                    reaction_line.append(str(abundances_metagenome[key][reaction]))
                                else:
                                    reaction_line.append(self.ZERO)

                            if(abundances_transcriptome and abundances_expression):
                                for key in self.metadata_keys:
                                    if reaction in abundances_transcriptome[key]:
                                        reaction_line.append(str(abundances_transcriptome[key][reaction]))
                                    else:
                                        reaction_line.append(self.ZERO)

                                for key in self.metadata_keys:
                                    if reaction in abundances_expression[key]:
                                        reaction_line.append(str(abundances_expression[key][reaction]))
                                    else:
                                        reaction_line.append(self.ZERO)

                            output_line = '\t'.join(reaction_line)
                            if output_line not in seen_steps:
                                seen_steps.add(output_line)
                                network_lines.append(output_line+'\t%i' % steps)
                                
                            #~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#~#
                            #~#~#~#~#~#~#~ Fill in node metadata  ~#~#~#~#~#~#~
                            index = str((steps if compound 
                                         in queries_list else steps+1))
                            
                            if compound not in seen_nodes:
                                #if compound in query_list:
                                #    query_ab = [str(query_list[compound][key]) 
                                #                for key in self.metadata_keys]
                                #else:
                                #    query_ab = ['NA' for key 
                                #                in self.metadata_keys]
                                
                                if compound in self.d.compound_desc_dict:
                                    compound_type = \
                                        ','.join(self.d.compound_desc_dict[compound]['A'])
                                else:
                                    compound_type = 'NA'
                                    
                                if compound in queries_list:
                                    is_query = 'True'
                                else:
                                    is_query = 'False'
                                    
                                node_metadata_lines.append(
                                    '\t'.join([compound, 
                                               self.d.c[compound],
                                               compound_type, 
                                               'NA',
                                               'NA', 
                                               'NA',
                                               'NA',  
                                               'compound'] +
                                               #query_ab +
                                               [is_query, index])
                                                           )
                                
                                seen_nodes.add(compound)
                                if compound not in queries_list:
                                    level_queries.add(compound)
                            if reaction not in seen_nodes:
                                module, module_description \
                                                = self._gather_module(reaction)
                                pathway, pathway_description \
                                                = self._gather_pathway(reaction)
                                node_metadata_lines.append(
                                    '\t'.join([reaction, 
                                               self.d.r[reaction],
                                               'NA', 
                                               module, 
                                               module_description, 
                                               pathway,
                                               pathway_description,
                                               'reaction',
                                               'False',
                                               index])
                                                           )
                                seen_nodes.add(reaction)

            steps+=1
            depth-=1
            
            logging.info("Step %i complete with %i queries to continue with" \
                                              % (steps, len(level_queries)))
        return network_lines, node_metadata_lines
    
    
    def pathway_matrix(self, abundances_metagenome,  abundances_transcriptome, abundances_expression, abundances_metabolome, limit, filter):

        
        possible_reactions=set()

        if any(limit):
            for entry in limit:
                if(entry.startswith(self.MAP_PREFIX) or 
                   entry.startswith(self.PATHWAY_PREFIX)): 
                    for reaction in self.d.p2r[entry]:
                        possible_reactions.add(reaction)
                elif(entry.startswith(self.MODULE_PREFIX)):
                    for reaction in self.d.m2r[entry]:
                        possible_reactions.add(reaction)
                elif(entry.startswith(self.REACTION_PREFIX)):
                    possible_reactions.add(entry)

            possible_reactions = {reaction:self.d.r2c[reaction] 
                                  for reaction in
                                  possible_reactions}                
        else:
            possible_reactions = self.d.r2c
        
        for entry in filter:
            if entry in possible_reactions:
                del possible_reactions[entry]
 

        network_lines, node_metadata_lines = \
            self.all_matrix(abundances_metagenome, 
                            abundances_transcriptome, 
                            abundances_expression, 
                            abundances_metabolome,
                            possible_reactions)
        
        return network_lines, node_metadata_lines
  
    def traverse(self, abundances_metagenome, abundances_transcriptome, limit, filter, 
                 starting_compounds, steps, number_of_queries):

        if any(limit):
            for entry in limit:
                if(entry.startswith(self.MAP_PREFIX) or 
                   entry.startswith(self.PATHWAY_PREFIX)): 
                    for reaction in self.d.p2r[entry]:
                        possible_reactions.add(reaction)
                elif(entry.startswith(self.MODULE_PREFIX)):
                    for reaction in self.d.m2r[entry]:
                        possible_reactions.add(reaction)
                elif(entry.startswith(self.REACTION_PREFIX)):
                    possible_reactions.add(entry)

            possible_reactions = {reaction:self.d.r2c[reaction] 
                                  for reaction in
                                  possible_reactions}                
        else:
            possible_reactions = self.d.r2c
        
        for entry in filter:
            if entry in possible_reactions:
                del possible_reactions[entry]
        
        possible_compounds = set(itertools.chain(*possible_reactions.values()))

        if len(starting_compounds)==0:
            query_list=[x for x in self.d.c.keys()
                        if x in possible_compounds]
        else:
            query_list=[x for x in starting_compounds 
                        if x in possible_compounds]

        output_lines = self.traverse_matrix(abundances_metagenome, 
                                            abundances_transcriptome, 
                                            possible_reactions, 
                                            query_list,
                                            number_of_queries, 
                                            steps)
        return output_lines


#!/usr/bin/env python
###############################################################################
#                                                                             #
#    This program is free software: you can redistribute it and/or modify     #
#    it under the terms of the GNU General Public License as published by     #
#    the Free Software Foundation, either version 3 of the License, or        #
#    (at your option) any later version.                                      #
#                                                                             #
#    This program is distributed in the hope that it will be useful,          #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of           #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            #
#    GNU General Public License for more details.                             #
#                                                                             #
#    You should have received a copy of the GNU General Public License        #
#    along with this program. If not, see <http://www.gnu.org/licenses/>.     #
#                                                                             #
###############################################################################

__author__ = "Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"

###############################################################################
# imports
import os
import logging
from Bio import SeqIO

###############################################################################

class Genome:
	'''
	A genome object which collects all the attirbutes of an imput genome,
	including protein sequences and their annotations
	'''
	def __init__(self, path):

		self.protein_ordered_dict = {}
		self.path = path
		self.name = os.path.split(os.path.splitext(path)[0])[1]
		self.sequences = {}
		
		for protein_count, protein in enumerate(SeqIO.parse(path, 'fasta')):
			sequence = Sequence(protein.description, protein.seq)
		 	self.sequences[protein.name] = sequence
			self.protein_ordered_dict[protein_count] = protein.name
			
	def add(self, annotations, evalue_cutoff, bitscore_cutoff, 
		percent_aln_query_cutoff, percent_aln_reference_cutoff, annotation_type):
		'''
		Adds a series of annotations to the proteins within a genome.

		Parameters
		----------
		annotations						- String. Path to file containing either blast or
										  domtblout hmmsearch results. 
		evalue_cutoff					- Float. E-value threshold for annotations.
		bitscore_cutoff					- Float. Bit score threshold for annotations.
		percent_aln_query_cutoff		- Float. Threshold for the percent of the query 
										  that must be aligned to consider the annotation.
		percent_aln_reference_cutoff	- Float. Threshold for the percent of the reference
										  that must be aligned to consider the annotation.
		annotation_type					- String. Either 'KO', 'TIGRFAM', 'PFAM' or 'COG'

		'''
			
		# Load up annotation parser, and tell it what annotation type to expect
		ap = AnnotationParser(annotation_type)

		# If annotation type is a hmmsearch result
		if(annotation_type == AnnotationParser.PFAM or
		   annotation_type == AnnotationParser.TIGRFAM):
			# Set up an iterator to produce the results
			iterator = ap.from_hmmsearch_results(annotations, evalue_cutoff,
												 bitscore_cutoff, percent_aln_query_cutoff, 
												 percent_aln_reference_cutoff)

		# If annotation type is a blast result
		elif(annotation_type == AnnotationParser.KO or
			 annotation_type == AnnotationParser.COG):
			# Set up an iterator to produce the results
			iterator = ap.from_blast_results(annotations, evalue_cutoff, 
											 bitscore_cutoff, percent_aln_query_cutoff)


		for seqname, annotation, evalue, annotation_range in iterator:
			self.sequences[seqname].add(annotation, evalue, annotation_range, annotation_type)		

	def count(self, annotation):
		'''
		
		Parameters
		----------
		annotation - String. An annotation ID to return a frequency for
		
		Output
		------
		The number of times this annotation was encountered in the genome
		'''
		count = 0
		for sequence in self.sequences.values():
			if annotation in sequence.all_annotations():
				count+=1
		return count 
	
	def ordered_sequences(self):
		'''
		Iterator that yields that all protein coding Sequence objects in a genome in order.
		'''
		for sequence_id in sorted(self.protein_ordered_dict.keys()):
			yield self.sequences[self.protein_ordered_dict[sequence_id]]

class Sequence(Genome):
	'''
	Sequence object which collects all attributes of a sequence including its length,
	and annotations. Can compare current annotation with new annotaitons.
	'''
	def __init__(self, description, sequence):
		self.annotations = []	
		self.seq = str(sequence)
		self.length = int(len(sequence))
		try:
			self.seqname, self.startpos, self.finishpos, self.direction, stats \
								= description.split(' # ')
			self.prod_id, self.partial, self.starttype, self.rbs_motif, self.rbs_spacer,  self.gc \
								= [x.split('=')[1] for x in stats.split(';')]
		except:
			raise Exception("Error parsing genome proteins. Was the output from prodigal?")

	def all_annotations(self):
		'''
		Returns a list of all annotations assigned to this sequence
		'''
		result = []
		for annotation in self.annotations:
			result.append(annotation.annotation)
		return result

	def seqdict(self):
		'''
		Output
		------
		A dictionary where each entry is a position in the sequence,
		and values are the annotation at that position. This is important
		in particular for pfam annotations as a protein can have >1 
		domains.
		'''
		seq_dict = {x:None for x in range(self.length)}
		for annotation in self.annotations:
			for position in annotation.region:
				seq_dict[position] = annotation.annotation
		return seq_dict


	def what(self, query_region):
		'''
		Return annotations assigned to a list of positions within a sequence.

		Parameters
		----------
		region 		- list. List of integers specifying the positions in the 
					  sequence to return annotations for
		Outputs
		-------
		Returns a list of equal length to region, containing the annotation of
		each position.
		'''
		result = []

		# Build reference dictionary for sequence
		seq_dict = self.seqdict()

		# Find annotation for each position in the sequence
		for position in query_region:
			result.append(seq_dict[position])
		return result

	def add(self, annotation, evalue, region, annotation_type):
		'''
		Return annotations assigned to a list of positions within a sequence.

		Parameters
		----------
		annotation 	- string. Annotation to assign to the sequence region
		evalue 		- float. Evalue awarded for the given annotation
		region		- list. List of integers specifying the positions in the 
					  sequence to annotate
		'''
		new_annotation = Annotation(annotation, evalue, region, annotation_type)
		if len([x for x in self.annotations if x.type == new_annotation.type]) > 0:
			for idx, previous_annotation in enumerate(self.annotations):
				if previous_annotation.type == new_annotation.type:
					if len(previous_annotation.region.intersection(new_annotation.region)) > 0:
						is_better = new_annotation.compare(previous_annotation)
						if is_better:
							self.annotations[idx] = new_annotation
		else:
			self.annotations.append(new_annotation)

	
class Annotation(Sequence):
	'''
	Annotation object that collects all attributes assocaited with a given
	annotation, like where it is in the sequence, its e-value and bit score,
	and what type of annotation it is.
	'''
	def __init__(self, annotation, evalue, region, annotation_type):

		self.annotation = annotation
		self.evalue 	= float(evalue)
		self.region		= set(region)
		self.type   	= annotation_type

	def compare(self, other_annotation):
		'''
		Compares evalue of current annotation with the evaulue of another annotation object. 			

		Parameters
		----------
		other_annotation: Annotation object.
		
		Output
		------
		Returns True if self is the better annotation or False if it isn't
		'''
		
		if self.evalue < other_annotation.evalue:
			return True
		else:
			return False

class AnnotationParser:
	'''
	Annotation parser class contains fucntions to parse hmmsearch domtblouts and blast results 
	currently for: KO, PFAM and TIGRFAM. COG to come	
	'''
	KO      	= 'KO_IDS.txt'
	PFAM    	= 'PFAM_IDS.txt'
	TIGRFAM 	= 'TIGRFAM_IDS.txt'
	COG 		= None ### ~ TODO: Not currently implemented

	def __init__(self, annotation_type):        
		

		data_directory = os.path.join(os.path.split(os.path.realpath(__file__))[0], '../data/ids/')
		if annotation_type == self.KO:
			ids = [x.strip() for x in open(os.path.join(data_directory,self.KO))]
		elif annotation_type == self.PFAM:
			ids = [x.strip() for x in open(os.path.join(data_directory,self.PFAM))]
		elif annotation_type == self.TIGRFAM:
			ids = [x.strip() for x in open(os.path.join(data_directory,self.TIGRFAM))]

	def from_blast_results(self,
						   blast_output_path,
						   evalue_cutoff,
						   bitscore_cutoff, 
						   percent_id_cutoff):
		'''
		Parse blast output in tab format.

		Parameters
		----------
		blast_output_path 	- String. Path to blast output file containing results.
							  Must be in blast output format 6
		evalue_cutoff		- Float. E-value threshold for annotations.
		bitscore_cutoff		- Float. Bit score threshold for annotations.
		percent_id_cutoff 	- Float. Percent identity threshold for annotations.
		
		Yields
		------
		A sequence name, annotation, E-value and region hit for every annottation result in 
		blast_output_path that pass a series of specified cutoffs
		'''

		logging.info("Parsing blast output file: %s" % blast_output_path)

		for line in open(blast_output_path):
			# Parse out important information from each line in blast output
			sline    = line.strip().split()
			evalue   = sline[10]
			bit      = sline[11]
			perc_id  = sline[2]
			seq_list = [int(x) for x in sline[6:8]]

			# If the annotation passes the specified cutoffs
			if(float(evalue) <= evalue_cutoff and
				float(bit) >= bitscore_cutoff and
				float(perc_id)  >= percent_id_cutoff):
					seqname = sline[0]
					annotation = sline[1].split('~')[1]
					yield seqname, annotation, evalue, range(min(seq_list), max(seq_list))

	def from_hmmsearch_results(self,
							   hmmsearch_output_path,
							   evalue_cutoff,
							   bitscore_cutoff, 
    						   percent_aln_query_cutoff,
    						   percent_aln_reference_cutoff):
		'''
		Parse input hmmsearch file

		Parameters 
		----------
		hmmsearch_output_path           - String. Path to domtblout file containing 
										  hmmsearch results.
		evalue_cutoff                   - Float. E-value threshold for annotations.
		bitscore_cutoff                 - Float. Bit score threshold for annotations.
		percent_aln_query_cutoff        - Float. Threshold for the percent of the query 
										  that must be aligned to consider the annotation.
		percent_aln_reference_cutoff    - Float. Threshold for the percent of the reference 
										  that must be aligned to consider the annotation.
		Yields
		------
		A sequence name, accession, E-value and region hit for every annottation result in 
		blast_output_path that pass a set of cutoffs
		'''

		logging.debug("    - Parsing hmmsearch output file: %s" % hmmsearch_output_path)

		# Filling in column
		for line in open(hmmsearch_output_path):
			
			# Skip headers
			if line.startswith('#'): continue
				
			# Parse HMMsearch line. '_'s represent unimportant entries. Line
			# is trimmed using [:22] to remove sequence description
			seqname, _, tlen, annotation, accession, qlen, evalue, score, \
			bias, _, _, c_evalue, i_evalue, dom_score, dom_bias, hmm_from, \
			hmm_to, seq_from, seq_to, _, _, acc = line.strip().split()[:22]				

			# Determine sequence and HMM spans
			seq_list = [int(seq_from), int(seq_to)]
			hmm_list = [int(hmm_from), int(hmm_to)]

			# Calculate percent of the query and reference aligned to each-other. 
			perc_seq_aln = (max(seq_list)-min(seq_list))/float(tlen)
			perc_hmm_aln = (max(seq_list)-min(seq_list))/float(qlen)

			# If the annotation passes the specified cutoffs
			if(float(evalue)<=evalue_cutoff and
				float(score)>=bitscore_cutoff and
				perc_seq_aln>=percent_aln_query_cutoff and
				perc_hmm_aln>=percent_aln_reference_cutoff):
				
				yield seqname, accession, evalue, range(min(seq_list), max(seq_list))


#!/usr/bin/env python2
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
 
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2015"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.1"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"
 
###############################################################################

import logging
import os
import gzip

from genome import Genome

###############################################################################

class MatrixGenerator:
    
    COMPRESSED_SUFFIXES = set(['.gz', '.gzip'])
    REFERENCE_PATH = '/srv/db/uniprot/201607/KO.idmapping.dat.gz' ### ~ TODO: Wrap this up
    OLD_REFERENCE_PATH='/srv/db/uniprot/uniref_20151020/idmapping.KO.dat.gz'
    MATRIX_SUFFIX  = '_matrix.tsv'
    UR100 = 'UniRef100_'
    
    KO      = 'KO_IDS.txt'
    PFAM    = 'PFAM_IDS.txt'
    TIGRFAM = 'TIGRFAM_IDS.txt'

    def __init__(self, annotation_type):        
        data_directory = os.path.join(os.path.split(os.path.realpath(__file__))[0], '../data/ids/')
        if annotation_type == self.KO:
            self.annotation_list = [x.strip() for x in open(os.path.join(data_directory, self.KO))]
        elif annotation_type == self.PFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(data_directory, self.PFAM))]
        elif annotation_type == self.TIGRFAM:
            self.annotation_list = [x.strip() for x in open(os.path.join(data_directory, self.TIGRFAM))]

    def write_matrix(self, genomes_list, output_path):
        '''
        Writes a frequency matrix with of each annotation (rows) per sample (columns)

        Parameters
        ----------
        genomes_list        - list. List of Genome objects
        output_path         - string. Path to file to which the results are written.
        '''
        logging.info("    - Writing results to file: %s" % output_path)
        with open(output_path, 'w') as out_io:
            colnames = ['ID'] + [genome.name for genome in genomes_list]
            out_io.write('\t'.join(colnames) + '\n')
            for annotation in self.annotation_list:
                output_line = [annotation]
                for genome in genomes_list:
                    output_line.append( str(genome.count(annotation)) )
                out_io.write( '\t'.join(output_line) + '\n' )



#!/usr/bin/env python
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################

__author__ = "Ben Woodcroft, Joel Boyd"
__copyright__ = "Copyright 2017"
__credits__ = ["Ben Woodcroft"]
__license__ = "GPL3+"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd3 near uq.edu.au"
__status__ = "Development"

###############################################################################
# Imports

import re

###############################################################################

class ModuleDescription:

    def __init__(self, module_description_string):
        self.module_description_string = module_description_string
        self.parsed_module = ModuleDescriptionParser().parse_module_string(
            module_description_string)

        if isinstance(self.parsed_module, ModuleDescriptionOrRelation):
            new_and = ModuleDescriptionAndRelation()
            new_and.relations = [self.parsed_module]
            self.parsed_module = new_and

    def amount_of_pathway_covered(self, ko_list):
        pass #FIXME self._global_steps = module_description_string.split(' ')

    def kos(self):
        '''Return an iterable over the total list of KOs in the module'''
        r = re.compile('(K\d+)')
        return re.findall(r, self.module_description_string)

    def num_steps(self):
        if isinstance(self.parsed_module, ModuleDescriptionAndRelation):
            return len(self.parsed_module.relations)
        else:
            raise Exception("Cannot work with non-AND type modules")

    def num_covered_steps(self, ko_set):
        if isinstance(self.parsed_module, ModuleDescriptionAndRelation):
            step_cov = 0
            path_cov = 0 
            reac_cov = 0
            for m in self.parsed_module.relations:
                step_passed, step_counts, reaction_counts = m.satisfied_with(ko_set)
                if step_passed:
                    step_cov+=1
                    path_cov+=step_counts
                reac_cov+=reaction_counts
            return step_cov, path_cov, reac_cov
        
        else:
            raise Exception("Cannot work with non-AND type modules")

class ModuleDescriptionAndRelation:
    def satisfied_with(self, set_of_kos):
        
        counts          = 0
        step_passed     = False
        reaction_counts = 0
        founds          = []
        
        for r in self.relations:
            found, count, reaction_count = r.satisfied_with(set_of_kos)
            if found:
                founds.append(1)
                counts += count
            reaction_counts+=reaction_count
        step_passed = len(self.relations) == sum(founds)
        
        return step_passed, counts, reaction_counts

class ModuleDescriptionOrRelation:
    def satisfied_with(self, set_of_kos):

        counts          = 0
        step_passed     = False
        reaction_counts = 0

        for r in self.relations:
            found, count, reaction_count = r.satisfied_with(set_of_kos)
            if found:
                step_passed = True
                return step_passed, counts, reaction_count
            reaction_counts+=reaction_count
        return step_passed, counts, reaction_counts

class ModuleDescriptionPlusRelation(ModuleDescriptionAndRelation): pass

class ModuleDescriptionKoEntry:
    def __init__(self, ko):
        self.ko = ko

    def satisfied_with(self, set_of_kos):
        found = self.ko in set_of_kos
        count = (1 if found else 0)
        reaction_count = 1
        return found, count, reaction_count

class ParserHelper: pass

class ModuleDescriptionParser:
    
    def correct_substrings(self, substring_list):
        fixed_substrings = []
        for substring in substring_list:
            # Omit optional enzyme (e.g. M00372) 
            # or undefined KO group (e.g. M00079)
            if substring.startswith('-'):
                continue 
            # Remove redundant and definitions
            substring=substring.replace(', ', ',')
            fixed_substrings.append(substring)
        return fixed_substrings
        
    def parse_module_string(self, string):

        frags1 = self.split_on_space(string)
        frags1 = self.correct_substrings(frags1) 

        if len(frags1) == 1:
            # rare if ever, I think
            frags1 = self.split_on_comma(frags1[0])
            master_relation = ModuleDescriptionOrRelation()
        else:
            master_relation = ModuleDescriptionAndRelation()
            
        current = ParserHelper()
        current.top_relation = master_relation
        current.understuff = frags1
        stack = list([current])
        while len(stack) > 0:
            current = stack.pop()
            new_stuff = []
            for e in current.understuff:
                if isinstance(e, str):
                    if re.match('^K\d+$', e):
                        new_stuff.append(ModuleDescriptionKoEntry(e))
                    else:
                        #TOREMOVE frags = self.split_on_space(e)
                        frags = self.split_on_comma(e)
                        if len(frags) == 1:
                            topush = ParserHelper()
                            #TOREMOVE comma_splits = self.split_on_comma(e)
                            comma_splits = self.split_on_space(e)
                            m = None
                            if len(comma_splits) == 1:
                                plus_splits = self.split_on_plus(e)
                                minus_splits = self.split_on_minus(e)
                                if len(plus_splits)>1:
                                    m = ModuleDescriptionPlusRelation()
                                    topush.understuff = plus_splits
                                elif len(minus_splits)>1:
                                    m = ModuleDescriptionAndRelation()
                                    topush.understuff = minus_splits[:1]
                                else:
                                    raise Exception("Parse exception on %s" % string)
                            else:
                                #m = ModuleDescriptionOrRelation()
                                m = ModuleDescriptionAndRelation()
                                topush.understuff = comma_splits
                            topush.top_relation = m
                            stack.append(topush)
                            new_stuff.append(m)
                        else:
                            #m = ModuleDescriptionAndRelation()                            
                            m = ModuleDescriptionOrRelation()
                            topush = ParserHelper()
                            topush.top_relation = m
                            topush.understuff = frags
                            stack.append(topush)
                            new_stuff.append(m)
                else:
                    new_stuff.append(e)
            current.top_relation.relations = new_stuff
        return master_relation


    def split_on(self, string, characters):
        bracket_counter = 0
        fragments = []
        current = []
        remove_end_brackets = True
        for i in xrange(len(string)):
            c = string[i]
            if c == '(':
                current += c
                bracket_counter += 1
            elif c == ')':
                current += c
                bracket_counter -= 1
                if bracket_counter == 0 and i < len(string)-1:
                    remove_end_brackets = False
            elif c in characters and bracket_counter == 0:
                fragments.append(''.join(current))
                current = ''
            else:
                current += c
        fragments.append(''.join(current))

        if remove_end_brackets and string[0] == '(':
            if string[-1] != ')': raise Exception("Parse error")
            return self.split_on(string[1:-1], characters)
        return fragments

    def split_on_space(self, string):
        return(self.split_on(string, ' '))

    def split_on_plus(self, string):
        return(self.split_on(string, '+'))

    def split_on_comma(self, string):
        return(self.split_on(string, ','))

    def split_on_minus(self, string):
        return(self.split_on(string, '-'))


#!/usr/bin/env python
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################
 
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2015"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.1"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"
 
###############################################################################

import logging
import pickle
import os

###############################################################################

class KeggMatrix:
    
    DATA_PATH = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                             '..', 
                             'data')
    VERSION   = os.path.join(DATA_PATH, 'VERSION')
    R2K       = os.path.join(DATA_PATH, 'reaction_to_orthology')
    PICKLE    = 'pickle'
    
    def __init__(self, matrix, transcriptome):
        self.VERSION = open(self.VERSION).readline().strip()

        logging.info("Loading reaction to pathway information")
        self.r2k = pickle.load(open('.'.join([self.R2K, self.VERSION, 
                                              self.PICKLE])))
        logging.info("Done")
        logging.info("Parsing input matrix: %s" % matrix)
        self.orthology_matrix \
                 = self._parse_matrix(matrix)
        logging.info("Done")
        logging.info("Calculating reaction abundances")
        self.reaction_matrix  \
                 = self._calculate_abundances(self.r2k, self.orthology_matrix)
        logging.info("Done")
        
        if transcriptome:
            logging.info("Parsing input transcriptome: %s" % transcriptome)
            self.orthology_matrix_transcriptome \
                        = self._parse_matrix(transcriptome)
            logging.info("Done")
            
            logging.info("Calculating reaction transcriptome abundances")
            self.reaction_matrix_transcriptome \
                 = self._calculate_abundances(self.r2k, 
                                              self.orthology_matrix_transcriptome)
            logging.info("Done")
            
            logging.info("Calculating normalized expression abundances")
            self.orthology_matrix_expression \
                = self._calculate_expression_matrix(self.orthology_matrix, 
                                        self.orthology_matrix_transcriptome)
            logging.info("Done")
            
            logging.info("Calculating reaction expression abundances")
            self.reaction_matrix_expression  \
                 = self._calculate_abundances(self.r2k, 
                                              self.orthology_matrix_expression)
            logging.info("Done")
   
    def _calculate_expression_matrix(self, matrix, transcriptome_matrix):
        '''
        '''
        output_dictionary = {}
        for sample, abundances in transcriptome_matrix.items():
            output_dictionary[sample] = {}
            for ko, abundance in abundances.items():

                if ko in matrix[sample]:
                    mg_abundance = float(matrix[sample][ko])
                    mt_abundance = float(abundance)
                    if mg_abundance>0:
                        output_dictionary[sample][ko] \
                                = float(abundance)/float(matrix[sample][ko])
                else:
                    continue 
                    # for now ignore genes that are expressed, but not detected
                    # in the metagenome.
        return output_dictionary        
        
    def _parse_matrix(self, matrix):
        
        output_dict = {}
        
        for idx, line in enumerate(open(matrix)):
            sline = line.strip().split('\t')
            if idx==0:
                self.sample_names = sline[1:]
                    
                for sample in self.sample_names: 
                    output_dict[sample] = {}
            else:   
                ko_id      = sline[0]
                abundances = sline[1:]
                
                for abundance, sample in zip(abundances, self.sample_names):
                    output_dict[sample][ko_id] = float(abundance)
        return output_dict
    
    def group_abundances(self, samples, reference_dict):
        
        output_dict = {}        
        for sample in samples:
            new_dict = {key:entry for key,entry in reference_dict.items() if 
                        key in samples}
            reference_list = new_dict[new_dict.keys()[0]].keys()
            for reference in reference_list:
                # If samples are missing from stored data, this will crash 
                try:
                    abundances = [new_dict[sample][reference] for sample in samples]
                    average    = sum(abundances)/float(len(abundances))
                    if average > 0:
                        output_dict[reference] = average
                except:
                    raise Exception("metadata description does not match \
input matrix")
        return output_dict
                          
    def _calculate_abundances(self, reference_dict, matrix_dict):
        output_dict_mean   = {}
        for sample, ko_abundances in matrix_dict.items():
            output_dict_mean[sample]   = {}
            for reaction, ko_list in reference_dict.items():


                abundances = []
                for ko in ko_list:
                    if ko in matrix_dict[sample]:
                        abundances.append(matrix_dict[sample][ko])
                    else:
                        logging.debug("ID not found in input matrix: %s" % ko)
                if any(abundances):
                    abundance_mean = sum(abundances)/len(abundances)
                else:
                    abundance_mean = 0
                output_dict_mean[sample][reaction] = abundance_mean
        return output_dict_mean



from setuptools import setup, find_packages
with open('README.md') as readme_file:
    readme = readme_file.read()
exec(open('enrichm/version.py').read()) # loads __version__
setup(name='enrichm',
      version=__version__,
      author='Joel Boyd, Ben Woodcroft, Alex Baker',
      author_email='joel.boyd@uqconnect.edu.au',
      description='enrichm is a toolbox for comparing the functional composition of population genomes',
      long_description=readme,
      license='GPL3+',
      keywords=["MAGs", "Population genomes", "metagenomics", "Annotation", "Comparison"],
      packages=find_packages(exclude='docs'),
      install_requires=('python-dateutil >=2.5.1',
                        'statsmodels >=0.8.0rc1',
                        'numpy >=1.9.1',
                        'pandas >=0.17.1',
                        'scipy >=0.17.0',
                        'biopython >=1.66'),
      setup_requires=['nose>=1.0'],
      test_suite='nose.collector',
      url='https://github.com/geronimp/enrichM',
      download_url='https://github.com/geronimp/enrichm/archive/%s.tar.gz' % (__version__),
      scripts=['bin/enrichm'],
      data_files=[
          ('pickles', ['data/br08001.07-08-2017.pickle',
                       'data/clan_to_name.07-08-2017.pickle',
                       'data/clan_to_pfam.07-08-2017.pickle',
                       'data/compound_descriptions.07-08-2017.pickle',
                       'data/compound_to_reaction.07-08-2017.pickle',
                       'data/module_descriptions.07-08-2017.pickle',
                       'data/module_to_definition.07-08-2017.pickle',
                       'data/module_to_reaction.07-08-2017.pickle',
                       'data/pathway_descriptions.07-08-2017.pickle',
                       'data/pathway_to_reaction.07-08-2017.pickle',
                       'data/pfam_to_clan.07-08-2017.pickle',
                       'data/pfam_to_description.07-08-2017.pickle',
                       'data/pfam_to_name.07-08-2017.pickle',
                       'data/reaction_descriptions.07-08-2017.pickle',
                       'data/reaction_to_compound.07-08-2017.pickle',
                       'data/reaction_to_module.07-08-2017.pickle',
                       'data/reaction_to_orthology.07-08-2017.pickle',
                       'data/reaction_to_pathway.07-08-2017.pickle'],

           'ids',     ['data/ids/KO_IDS.txt',
                       'data/ids/PFAM_CLANS.txt',
                       'data/ids/PFAM_IDS.txt',
                       'data/ids/TIGRFAM_IDS.txt']),
      ],
)



#!/usr/bin/env python
###############################################################################
#                                                                             #
# This program is free software: you can redistribute it and/or modify        #
# it under the terms of the GNU General Public License as published by        #
# the Free Software Foundation, either version 3 of the License, or           #
# (at your option) any later version.                                         #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the                #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program. If not, see <http://www.gnu.org/licenses/>.        #
#                                                                             #
###############################################################################

 
__author__ = "Joel Boyd"
__copyright__ = "Copyright 2015"
__credits__ = ["Joel Boyd"]
__license__ = "GPL3"
__version__ = "0.0.1"
__maintainer__ = "Joel Boyd"
__email__ = "joel.boyd near uq.net.au"
__status__ = "Development"

###############################################################################

import logging
import math
import numpy as np

from kegg_module_grabber import KeggModuleGrabber
from module_description_parser import ModuleDescription

class MetagenomeAnalyzer:
    
    MODULE_AB = 'module_ab'
    KO_PREFIX = 'K'
    
    def __init__(self):
        self.module_ab_output_suffix = '_module_abundances.tsv'
    
    def _extract_kos(self, string):
        output_list = []
        for idx, char in enumerate(string):
            if char == self.KO_PREFIX:
                output_list.append(string[idx:idx+6])
                
        return output_list
    
    def calculate_module_abundance(self, matrix_path, module_to_def, 
                                   needed_kos=None):
        logging.info("Loading KO matrix: %s" % (matrix_path))
        ko_matrix_object = Matrix(matrix_path)
        mkg = KeggModuleGrabber()
        output_lines = ['\t'.join(["Module", "Sample", 
                                   "Mean", "SD", "SEM"])]
        needed_kos = (needed_kos if needed_kos else ko_matrix_object.rownames)

        logging.info("Generating list of possible KOs for each sample")
        sample_to_possible_kos = {sample:set() for sample in 
                                  ko_matrix_object.colnames}

        for ko in needed_kos:
            for sample in ko_matrix_object.colnames:                
                if float(ko_matrix_object.get_entry(sample, ko))>=0.0:       
                    sample_to_possible_kos[sample].add(ko)
        
        for module, definition in module_to_def.items():
            module_is_possible = False
            output_line_base = [module]
            sample_batches = []
            definition_object = ModuleDescription(definition)
            
            for sample in ko_matrix_object.colnames:
                ko_values = []
                possible_kos_in_sample = set()
                for ko in needed_kos:
                    if float(ko_matrix_object.get_entry(sample, ko))>=0.0:       
                        possible_kos_in_sample.add(ko)
                if(definition_object.num_steps()==definition_object.num_covered_steps(list(sample_to_possible_kos[sample]))):
                    module_is_possible=True
                    for ko in definition_object.kos():  
                        if ko in needed_kos:
                            ko_values.append(float(ko_matrix_object.get_entry(sample, ko)))
                    ko_values_array = np.array(ko_values)
                    mu=ko_values_array.mean()
                    sd=ko_values_array.std()
                    se=sd / math.sqrt(len(ko_values))
                    sample_batches.append([sample, str(mu), 
                                           str(sd), str(se)])
                    module_is_possible = True
                else:
                    sample_batches.append([sample, '0.0', '0.0', '0.0'])
            
            if module_is_possible:
                for batch in sample_batches:
                    line = output_line_base + batch
                    output_lines.append('\t'.join(line))
        return output_lines

    def main(self, args):
        if args.subparser_name == self.MODULE_AB:
            output_path = args.output_prefix + self.module_ab_output_suffix
            kmg = KeggModuleGrabber()
            if(args.modules or args.module_list_file):
                module_to_def = {}
                needed_kos=set()
                if args.modules:
                    for module in args.modules:
                        module_to_def[module] = kmg.m2def[module]
                        for ko in self._extract_kos(kmg.m2def[module]):
                            needed_kos.add(ko)
                elif args.module_list_file:
                    for module in [x.strip() 
                                   for x in open(args.module_list_file)]:
                        module_to_def[module] = kmg.m2def[module]
                        for ko in self._extract_kos(kmg.m2def[module]):
                            needed_kos.add(ko)
            else:
                module_to_def = kmg.m2def
                needed_kos=False
            
            output_lines = self.calculate_module_abundance(args.matrix,
                                                           module_to_def,
                                                           needed_kos)
            
            logging.info('Writing output to file: %s' % (output_path))
            with open(output_path, 'w') as out_io:
                out_io.write('\n'.join(output_lines) + '\n')
            
            
        